@@ -16,6 +16,7 @@ def calc_sensitivities(sim        : 'Sim',
                        outvarname : str,
                        Hj         : float = 1.0,
                        tol        : float = 1e-6,
+                       verbose    : bool  = False,
                        ) -> tuple[np.ndarray, np.ndarray]:
     """
     Calculates the global sensitivity indices and ratios for a specific output
@@ -44,6 +45,8 @@ def calc_sensitivities(sim        : 'Sim',
     tol : float, default 1e-6
         The convergence tolerance for scipy's minimize function acting on the
         negative log likelihood function.
+    verbose : bool, default False
+        Whether to print diagnostic information.
 
     Returns
     -------
@@ -65,12 +68,12 @@ def calc_sensitivities(sim        : 'Sim',
            comprehensive, robust, and efficient global sensitivity analysis:
            1. Theory." Water Resources Research 52.1 (2016): 423-439.
     """
-    phi_opt = calc_phi_opt(sim, outvarname, tol)
+    phi_opt = calc_phi_opt(sim, outvarname, tol, verbose)
 
     variance = np.var(np.array([sim.outvars[outvarname].nums]))
     sensitivities = np.zeros(sim.ninvars)
     for j in range(sim.ninvars):
-        sensitivities[j] = calc_Gammaj(Hj, phi_opt[j], variance)
+        sensitivities[j] = calc_Gammaj(Hj, phi_opt[j], variance)/variance
     ratios = sensitivities/sum(sensitivities)
 
     return sensitivities, ratios
@@ -79,6 +82,7 @@ def calc_sensitivities(sim        : 'Sim',
 def calc_phi_opt(sim        : 'Sim',
                  outvarname : str,
                  tol        : float = 1e-6,
+                 verbose    : bool = False
                  ) -> np.ndarray:
     """
     Calculate the optimal hyperparameters for the covariance functions between
@@ -96,6 +100,8 @@ def calc_phi_opt(sim        : 'Sim',
     tol : float, default 1e-6
         The convergence tolerance for scipy's minimize function acting on the
         negative log likelihood function.
+    verbose : bool, default False
+        Whether to print diagnostic information.
 
     Returns
     -------
@@ -115,7 +121,7 @@ def calc_phi_opt(sim        : 'Sim',
     vprint(sim.verbose, 'Calculating optimal hyperparameters Φ for ' +
                        f"'{outvarname}' covariances...")
     X, Y = full_states(sim, outvarname)
-    res = minimize(L_runner, phi0s, args=(X, Y, sim.verbose), bounds=bounds,
+    res = minimize(L_runner, phi0s, args=(X, Y, verbose), bounds=bounds,
                    tol=tol, method='L-BFGS-B')
     phi_opt = res.x
     vprint(sim.verbose, 'Done calculating optimal hyperparameters.')
@@ -220,7 +226,7 @@ def L_runner(phi     : np.ndarray,
     return L
 
 
-@jit(nopython=True)
+@jit(nopython=True, cache=True)
 def calc_L(phi : np.ndarray,
            X   : np.ndarray,
            Y   : np.ndarray
@@ -257,7 +263,7 @@ def calc_L(phi : np.ndarray,
     return L
 
 
-@jit(nopython=True)
+@jit(nopython=True, cache=True)
 def calc_R(phi : np.ndarray,
            X   : np.ndarray,
            ) -> np.ndarray:
@@ -289,7 +295,7 @@ def calc_R(phi : np.ndarray,
     return R
 
 
-@jit(nopython=True)
+@jit(nopython=True, cache=True)
 def calc_Ruw(phi : np.ndarray,
              Xu  : np.ndarray,
              Xw  : np.ndarray
@@ -319,7 +325,7 @@ def calc_Ruw(phi : np.ndarray,
     return Ruw
 
 
-@jit(nopython=True)
+@jit(nopython=True, cache=True)
 def calc_rj(hj   : float,
             phij : float
             ) -> float:
