@@ -4,20 +4,26 @@ from __future__ import annotations
 import os
 import numpy as np
 import dask
-from dask.distributed import Client
+import csv
+import json
 import cloudpickle
 import pathlib
+from dask.distributed import Client
 from datetime import datetime, timedelta
+from matplotlib.figure import Figure
+from matplotlib.axes import Axes
+from tqdm import tqdm
+from typing import Callable, Any, Iterable, Optional
+from scipy.stats import rv_continuous, rv_discrete
 from monaco.mc_case import Case
-from monaco.mc_var import InVar, OutVar
+from monaco.mc_var import InVar, OutVar, InVarSpace
 from monaco.mc_enums import SimFunctions, SampleMethod
-from monaco.helper_functions import get_list, vprint, vwarn, hash_str_repeatable
+from monaco.helper_functions import (get_list, vprint, vwarn, empty_list,
+                                     hash_str_repeatable)
 from monaco.case_runners import preprocess_case, run_case, postprocess_case
 from monaco.tqdm_dask_distributed import tqdm_dask
 from monaco.dvars_sensitivity import calc_sensitivities
-from tqdm import tqdm
-from typing import Callable, Any, Iterable
-from scipy.stats import rv_continuous, rv_discrete
+from monaco.mc_multi_plot import multi_plot_grid_rect
 
 
 class Sim:
@@ -51,6 +57,10 @@ class Sim:
     debug : bool, default: False
         If False, cases that fail while running will be skipped over. If True,
         cases that fail will raise an exception.
+    keepsiminput : bool, default: True
+        Whether to keep the siminput for each case after running.
+    keepsimrawoutput : bool, default: True
+        Whether to keep the simrawoutput for each case after postprocessing.
     savesimdata : bool, default: True
         Whether to save the simulation data to disk as a .mcsim file.
     savecasedata : bool, default: True
@@ -86,6 +96,8 @@ class Sim:
         The case numbers which were sucessfully run.
     casespostprocessed : set[int]
         The case numbers which were sucessfully postprocessed.
+    vars : dict[str, monaco.mc_var.InVar]
+        All Variables.
     invars : dict[str, monaco.mc_var.InVar]
         The Monte Carlo Input Variables.
     outvars : dict[str, monaco.mc_var.OutVar]
@@ -116,12 +128,14 @@ class Sim:
                  firstcaseismedian : bool = False,
                  samplemethod      : SampleMethod = SampleMethod.SOBOL_RANDOM,
                  seed              : int  = np.random.get_state(legacy=False)['state']['key'][0],
-                 singlethreaded    : bool = False,
+                 singlethreaded    : bool = True,
                  daskkwargs        : dict = dict(),
                  verbose           : bool = True,
                  debug             : bool = False,
-                 savesimdata       : bool = True,
-                 savecasedata      : bool = True,
+                 keepsiminput      : bool = True,
+                 keepsimrawoutput  : bool = True,
+                 savesimdata       : bool = False,
+                 savecasedata      : bool = False,
                  resultsdir        : str | pathlib.Path = None,
                  ) -> None:
 
@@ -137,6 +151,8 @@ class Sim:
         self.seed = seed
         self.singlethreaded = singlethreaded
         self.daskkwargs = daskkwargs
+        self.keepsiminput = keepsiminput
+        self.keepsimrawoutput = keepsimrawoutput
         self.savesimdata = savesimdata
         self.savecasedata = savecasedata
 
@@ -163,6 +179,7 @@ class Sim:
         self.casesrun           : set[int] = set()
         self.casespostprocessed : set[int] = set()
 
+        self.vars    : dict[str, InVar | OutVar] = dict()
         self.invars  : dict[str, InVar] = dict()
         self.outvars : dict[str, OutVar] = dict()
         self.constvals : dict[str, Any] = dict()
@@ -266,7 +283,7 @@ class Sim:
             nthreads = nworkers * self.cluster.worker_spec[0]['options']['nthreads']
             memory = nworkers * self.cluster.worker_spec[0]['options']['memory_limit']
             vprint(self.verbose,
-                   f'Dask cluster initiated with {nworkers} workers,' +
+                   f'Dask cluster initiated with {nworkers} workers, ' +
                    f'{nthreads} threads, {memory/2**30:0.2f} GiB memory.')
             vprint(self.verbose, f'Dask dashboard link: {self.cluster.dashboard_link}')
 
@@ -277,6 +294,7 @@ class Sim:
                  distkwargs : dict[str, Any],
                  nummap     : dict[float, Any] = None,
                  seed       : int = None,
+                 datasource : Optional[str] = None,
                  ) -> None:
         """
         Add an input variable to the simulation.
@@ -288,13 +306,20 @@ class Sim:
         dist : scipy.stats.rv_discrete | scipy.stats.rv_continuous
             The statistical distribution to draw from.
         distkwargs : dict
-            The keyword argument pairs for the statistical distribution function.
+            The keyword argument pairs for the statistical distribution
+            function.
         nummap : dict[float, Any], default: None
             A dictionary mapping numbers to nonnumeric values.
         seed : int
             The random seed for this variable. If None, a seed will be assigned
             based on the order added.
+        datasource : str, default: None
+            If the invals were imported from a file, this is the filepath. If
+            generated through monaco, then None.
         """
+        if name in self.vars.keys():
+            raise ValueError(f"'{name}' is already a Variable")
+
         self.ninvars += 1
         if seed is None:
             # seed is dependent on the order added
@@ -302,8 +327,10 @@ class Sim:
         self.invarseeds.append(seed)
         invar = InVar(name=name, dist=dist, distkwargs=distkwargs, ndraws=self.ndraws,
                       nummap=nummap, samplemethod=self.samplemethod, ninvar=self.ninvars,
-                      seed=seed, firstcaseismedian=self.firstcaseismedian, autodraw=False)
+                      seed=seed, firstcaseismedian=self.firstcaseismedian, autodraw=False,
+                      datasource=datasource)
         self.invars[name] = invar
+        self.vars[name] = invar
 
 
     def addConstVal(self,
@@ -347,9 +374,11 @@ class Sim:
         """Draw the random values for all the input variables."""
         if self.ninvars > 0:
             vprint(self.verbose, f"Drawing random samples for {self.ninvars} input variables " +
-                                 f"via the '{self.samplemethod}' method...", flush=True)
-        for invar in self.invars.values():
-            invar.draw(ninvar_max=self.ninvars)
+                                 f"via the '{self.samplemethod}' method...", end=' ', flush=True)
+            for invar in self.invars.values():
+                if invar.datasource is None:
+                    invar.draw(ninvar_max=self.ninvars)
+            vprint(self.verbose, 'Done', flush=True)
 
 
     def runSim(self,
@@ -481,7 +510,7 @@ class Sim:
             The case numbers to generate. If None, then all cases are
             generated.
         """
-        vprint(self.verbose, 'Generating cases...', flush=True)
+        vprint(self.verbose, 'Generating cases...', end=' ', flush=True)
         self.genCaseSeeds()
 
         # If we are rerunning partial cases we don't want to reset this
@@ -494,8 +523,11 @@ class Sim:
             if self.firstcaseismedian and ncase == 0:
                 ismedian = True
             self.cases.append(Case(ncase=ncase, ismedian=ismedian, invars=self.invars,
-                                   constvals=self.constvals, seed=int(self.caseseeds[ncase])))
+                                   constvals=self.constvals, keepsiminput=self.keepsiminput,
+                                   keepsimrawoutput=self.keepsimrawoutput,
+                                   seed=int(self.caseseeds[ncase])))
         self.cases.sort(key=lambda case: case.ncase)
+        vprint(self.verbose, 'Done', flush=True)
 
 
     def genCaseSeeds(self) -> None:
@@ -652,7 +684,7 @@ class Sim:
         # Single-threaded for loop
         if self.singlethreaded:
             if self.verbose:
-                pbar = tqdm(total=len(cases_downselect), desc='Preprocessing cases',
+                pbar = tqdm(total=len(cases_downselect), desc='Postprocessing cases',
                             unit=' cases', position=0)
             for case in self.cases:
                 if case.ncase in cases_downselect:
@@ -696,10 +728,23 @@ class Sim:
                 self.casespostprocessed.add(case.ncase)
 
 
-    def genOutVars(self) -> None:
-        """Generate the output variables."""
+    def genOutVars(self,
+                   datasource : Optional[str] = None,
+                   ) -> None:
+        """
+        Generate the output variables.
+
+        Parameters
+        ----------
+        datasource : str, default: None
+            If the outvals were imported from a file, this is the filepath. If
+            generated through monaco, then None.
+        """
         for i_var, varname in enumerate(self.cases[0].outvals.keys()):
-            # seed is dependent on the order added
+            if varname in self.invars.keys():
+                raise ValueError(f"'{varname}' is already a Variable")
+
+            # seed is dependent on the order added, used for bootstrapping
             seed = (self.seed - 1 - i_var) % 2**32
             self.outvarseeds.append(seed)
 
@@ -722,11 +767,14 @@ class Sim:
             else:
                 valmap = self.cases[0].outvals[varname].valmap
 
-            self.outvars[varname] = OutVar(name=varname, vals=vals, valmap=valmap,
-                                           ndraws=self.ndraws, seed=seed,
-                                           firstcaseismedian=self.firstcaseismedian)
+            outvar = OutVar(name=varname, vals=vals, valmap=valmap,
+                            ndraws=self.ndraws, seed=seed,
+                            firstcaseismedian=self.firstcaseismedian,
+                            datasource=datasource)
+            self.outvars[varname] = outvar
+            self.vars[varname] = outvar
             for i in range(self.ncases):
-                self.cases[i].outvars[varname] = self.outvars[varname]
+                self.cases[i].outvars[varname] = outvar
 
         self.noutvars = len(self.outvars)
 
@@ -745,7 +793,9 @@ class Sim:
 
 
     def calcSensitivities(self,
-                          outvarnames: None | str | Iterable[str] = None,
+                          outvarnames : None | str | Iterable[str] = None,
+                          tol         : float = 1e-6,
+                          verbose     : bool = False,
                           ) -> None:
         """
         Calculate the sensitivity indices for the specified outvars.
@@ -755,9 +805,14 @@ class Sim:
         outvarnames : None | str | Iterable[str] (default: None)
             The outvar names to calculate sensitivity indices for. If None,
             then calculates sensitivities for all scalar outvars.
+        tol : float, default 1e-6
+            The convergence tolerance for scipy's minimize function acting on the
+            negative log likelihood function.
+        verbose : bool, default False
+            Whether to print diagnostic information.
         """
         if outvarnames is None:
-            outvarname = list(self.scalarOutVars.keys())
+            outvarnames = list(self.scalarOutVars().keys())
         outvarnames = get_list(outvarnames)
 
         for outvarname in outvarnames:
@@ -765,8 +820,9 @@ class Sim:
                 vwarn(self.verbose, f"Output variable '{outvarname}' is not scalar," +
                                      'skipping sensitivity calculations.')
             else:
-                vprint(self.verbose, f"Calculating sensitivity indices for '{outvarname}'.")
-                sensitivities, ratios = calc_sensitivities(self, outvarname)
+                vprint(self.verbose, f"Calculating sensitivity indices for '{outvarname}'...")
+                sensitivities, ratios = calc_sensitivities(self, outvarname,
+                                                           tol=tol, verbose=verbose)
 
                 sensitivities_dict = dict()
                 ratios_dict = dict()
@@ -776,6 +832,7 @@ class Sim:
 
                 self.outvars[outvarname].sensitivity_indices = sensitivities_dict
                 self.outvars[outvarname].sensitivity_ratios = ratios_dict
+                vprint(self.verbose, "Done calculating sensitivity indices.")
 
 
     def genCovarianceMatrix(self) -> None:
@@ -785,14 +842,10 @@ class Sim:
         """
         self.covvarlist = []
         allnums = []
-        for var in self.invars.keys():
-            if self.invars[var].isscalar:
-                allnums.append(self.invars[var].nums)
-                self.covvarlist.append(self.invars[var].name)
-        for var in self.outvars.keys():
-            if self.outvars[var].isscalar:
-                allnums.append(self.outvars[var].nums)
-                self.covvarlist.append(self.outvars[var].name)
+        for var in self.vars.keys():
+            if self.vars[var].isscalar:
+                allnums.append(self.vars[var].nums)
+                self.covvarlist.append(self.vars[var].name)
         self.covs = np.cov(np.array(allnums))
         self.corrcoeffs = np.corrcoef(np.array(allnums))
 
@@ -833,8 +886,73 @@ class Sim:
         return self.covs, self.covvarlist
 
 
+    def plot(self,
+             scalarvars  : Optional[list[InVar | OutVar]] = None,
+             cases           : None | int | Iterable[int] = None,
+             highlight_cases : None | int | Iterable[int] = empty_list(),
+             rug_plot    : bool   = False,
+             cov_plot    : bool   = False,
+             cov_p       : None | float | Iterable[float] = None,
+             invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
+             fig         : Figure = None,
+             title       : str    = '',
+             plotkwargs  : dict   = dict(),
+             ) -> tuple[Figure, tuple[Axes, ...]]:
+        """
+        Plot all the scalar variables against each other in a grid.
+
+        Parameters
+        ----------
+        scalarvars : list[monaco.mc_var.InVar | monaco.mc_var.OutVar]
+            The variables to plot. If None, then grabs all the input variables
+            and scalar output variables.
+        cases : None | int | Iterable[int], default: None
+            The cases to plot. If None, then all cases are plotted.
+        highlight_cases : None | int | Iterable[int], default: []
+            The cases to highlight. If [], then no cases are highlighted.
+        rug_plot : bool, default: False
+            Whether to plot rug marks.
+        cov_plot : bool, default: False
+            Whether to plot a covariance ellipse at a certain gaussian percentile
+            level.
+        cov_p : None | float | Iterable[float], default: None
+            The gaussian percentiles for the covariance plot.
+        invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
+            The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
+            specifies this individually for each of varx, vary, and varz.
+        fig : matplotlib.figure.Figure, default: None
+            The figure handle to plot in. If None, a new figure is created.
+        title : str, default: ''
+            The figure title.
+
+        Returns
+        -------
+        (fig, axes) : (matplotlib.figure.Figure, (matplotlib.axes.Axes, ...))
+            fig is the figure handle for the plot.
+            axes is a tuple of the axes handles for the plots.
+        """
+        if scalarvars is None:
+            scalarvars = []
+            for invar in self.invars.values():
+                scalarvars.append(invar)
+            for scalaroutvar in self.scalarOutVars().values():
+                scalarvars.append(scalaroutvar)
+
+        fig, axs = multi_plot_grid_rect(varsx=list(self.invars.values()),
+                                        varsy=list(self.scalarOutVars().values()),
+                                        cases=cases, highlight_cases=highlight_cases,
+                                        rug_plot=rug_plot,
+                                        cov_plot=cov_plot, cov_p=cov_p,
+                                        invar_space=invar_space,
+                                        fig=fig, title=title, plotkwargs=plotkwargs)
+        return fig, axs
+
+
     def clearResults(self) -> None:
         """Clear all the simulation results."""
+        for varname in self.outvars.keys():
+            if varname in self.vars.keys():
+                del self.vars[varname]
         self.cases = []
         self.outvars = dict()
         self.casespreprocessed = set()
@@ -851,6 +969,7 @@ class Sim:
     def reset(self) -> None:
         """Completely reset the simulation to the default object state."""
         self.clearResults()
+        self.vars = dict()
         self.invars = dict()
         self.constvals = dict()
         self.ninvars = 0
@@ -896,6 +1015,224 @@ class Sim:
         return allCases
 
 
+    def exportVars(self,
+                   vars : dict[str, InVar | OutVar],
+                   filename : Optional[str | pathlib.Path],
+                   ) -> pathlib.Path:
+        """
+        Export the nums for the selected to file for use externally.
+
+        Parameters
+        ----------
+        vars : dict[str, InVar | OutVar]
+            The vars to save to file.
+        filename : Optional[str | pathlib.Path]
+            The file to save to. Must be a csv or json.
+            If a str, then will save in the resultsdir.
+
+        Returns
+        -------
+        filepath : pathlib.Path
+            The filepath the vars were saved to.
+        """
+        if vars == []:
+            raise ValueError('No vars to save to file.')
+
+        if isinstance(filename, str):
+            filepath = self.resultsdir / filename
+        elif isinstance(filename, pathlib.Path):
+            filepath = filename
+
+        if filepath.suffix.lower() not in ('.csv', '.json'):
+            raise ValueError(f"'{filename}' must be a .csv or .json file.")
+        if filepath.exists():
+            vwarn(self.verbose, f'{filepath.name} already exists, overwriting.')
+
+        if filepath.suffix.lower() == '.csv':
+            varnames = list(vars.keys())
+            for i, var in enumerate(vars.values()):
+                if i == 0:
+                    data_csv = np.array(var.nums)
+                else:
+                    data_csv = np.vstack([data_csv, np.array(var.nums)])
+
+            with open(filepath, 'w') as f:
+                writer = csv.writer(f, quoting=csv.QUOTE_NONNUMERIC)
+                writer.writerow(varnames)
+                for i in range(self.ncases):
+                    if len(data_csv.shape) == 1:
+                        writer.writerow((data_csv[i], ))
+                    else:
+                        writer.writerow(data_csv[:, i])
+
+        elif filepath.suffix.lower() == '.json':
+            data_json = dict()
+            for varname, var in vars.items():
+                data_json[varname] = np.array(var.nums).tolist()
+
+            with open(filepath, 'w') as f:
+                json.dump(data_json, f, indent=0)
+
+        return filepath
+
+
+    def exportInVars(self,
+                     filename : Optional[str | pathlib.Path] = None,
+                     ) -> None:
+        """
+        Export the drawn nums for all the invars to file for use externally.
+
+        Parameters
+        ----------
+        filename : Optional[str | pathlib.Path]
+            The file to save to. Must be a csv or json.
+            If a str, then will save in the resultsdir.
+            If None, then will save to '{self.name}_invarnums.json'.
+        """
+        vprint(self.verbose, 'Exporting InVar draws to file...', flush=True)
+
+        if filename is None:
+            filename = self.resultsdir / f'{self.name}_invarnums.json'
+
+        filepath = self.exportVars(self.invars, filename)
+
+        vprint(self.verbose, f"InVar nums saved in '{filepath.name}'", flush=True)
+
+
+    def exportOutVars(self,
+                      filename : Optional[str | pathlib.Path] = None,
+                      ) -> None:
+        """
+        Export the nums for all the outvars to file for use externally.
+
+        Parameters
+        ----------
+        filename : Optional[str | pathlib.Path]
+            The file to save to. Must be a csv or json.
+            If a str, then will save in the resultsdir.
+            If None, then will save to '{self.name}_outvarnums.json'.
+        """
+        vprint(self.verbose, 'Exporting InVar draws to file...', flush=True)
+
+        if filename is None:
+            filename = self.resultsdir / f'{self.name}_outvarnums.json'
+
+        filepath = self.exportVars(self.outvars, filename)
+
+        vprint(self.verbose, f"OutVar nums saved in '{filepath.name}'", flush=True)
+
+
+    def importVars(self,
+                   filepath : str | pathlib.Path,
+                   ) -> tuple[dict[str, list[Any]], pathlib.Path]:
+        """
+        Import values from an external file.
+
+        Parameters
+        ----------
+        filepath : str | pathlib.Path
+            The file to load from. Must be a csv or json.
+
+        Returns
+        -------
+        data : dict[str, Any]
+            A dictionary where the keys are the variable name and the values
+            are a list of the vals.
+        filepath : pathlib.Path
+            The full filepath for the file which was loaded from.
+        """
+        if isinstance(filepath, str):
+            filepath = pathlib.Path(filepath)
+
+        if filepath.suffix.lower() not in ('.csv', '.json'):
+            raise ValueError(f"'{filepath.name}' must be a .csv or .json file.")
+
+        if self.cases == []:
+            self.genCases()
+
+        data : dict[str, list] = dict()
+        if filepath.suffix.lower() == '.csv':
+            with open(filepath, 'r') as f:
+                reader = csv.reader(f, quoting=csv.QUOTE_NONNUMERIC)
+                headers = next(reader)
+                for outvalname in headers:
+                    data[outvalname] = []
+                for row in reader:
+                    for outvalname, vals in zip(headers, row):
+                        data[outvalname].append(vals)
+
+        elif filepath.suffix.lower() == '.json':
+            with open(filepath, 'r') as f:
+                data = json.load(f)
+
+        for vals_list in data.values():
+            if len(vals_list) != self.ncases:
+                raise ValueError(f'Length of data ({len(vals_list)}) must match ' +
+                                 f'the number of sim cases ({self.ncases}).')
+
+        return data, filepath
+
+
+    def importInVars(self,
+                     filepath : str | pathlib.Path,
+                     nummap   : dict[Any, float] = None,
+                     ) -> None:
+        """
+        Import draws from an external file as InVals.
+
+        Parameters
+        ----------
+        filepath : str | pathlib.Path
+            The file to load from. Must be a csv or json.
+        nummap : dict[Any, float], default: None
+            A nummap dict mapping numbers to nonnumeric values.
+        """
+        vprint(self.verbose, 'Importing InVals from file...', flush=True)
+
+        data, filepath = self.importVars(filepath)
+
+        pcts = [None for _ in range(self.ncases)]
+        for valname, nums in data.items():
+            self.addInVar(name=valname, dist=None, distkwargs=dict(), nummap=nummap,
+                          seed=None, datasource=str(filepath.resolve()))
+            nums = [np.array(num) for num in nums]
+            self.invars[valname].nums = nums
+            self.invars[valname].pcts = pcts
+            self.invars[valname].mapNums()
+
+
+        vprint(self.verbose, f"InVals loaded from '{filepath.name}' and converted to variables",
+               flush=True)
+
+
+    def importOutVars(self,
+                      filepath : str | pathlib.Path,
+                      valmap   : dict[Any, float] = None,
+                      ) -> None:
+        """
+        Import results from an external file as OutVals, convert to OutVars.
+
+        Parameters
+        ----------
+        filepath : str | pathlib.Path
+            The file to load from. Must be a csv or json.
+        valmap : dict[Any, float], default: None
+            A valmap dict mapping nonnumeric values to numbers.
+        """
+        vprint(self.verbose, 'Importing OutVals from file...', flush=True)
+
+        data, filepath = self.importVars(filepath)
+
+        for case in self.cases:
+            for valname, vals in data.items():
+                case.addOutVal(valname, vals[case.ncase], valmap=valmap)
+
+        self.genOutVars(datasource=str(filepath.resolve()))
+
+        vprint(self.verbose, f"OutVals loaded from '{filepath.name}' and converted to variables",
+               flush=True)
+
+
     def saveSimToFile(self) -> None:
         """Save the simulation to a .mcsim file"""
         if self.savesimdata:
