@@ -101,6 +101,11 @@ class InVal(Val):
         self.genValMap()
 
 
+    def __repr__(self):
+        return (f"{self.__class__.__name__}('{self.name}', ncase={self.ncase}, " +
+                f"val={self.val} ({self.num}), pct={self.pct:0.4f})")
+
+
     def mapNum(self) -> None:
         """
         Generate `val` based on the drawn number and the nummap.
@@ -177,6 +182,11 @@ class OutVal(Val):
         self.genNumMap()
 
 
+    def __repr__(self):
+        return (f"{self.__class__.__name__}('{self.name}', ncase={self.ncase}, " +
+                f"val={self.val} ({self.num}))")
+
+
     def convertPandas(self) -> None:
         """
         If the output value is a pandas dataseries or index, convert it to a
@@ -259,5 +269,5 @@ class OutVal(Val):
             for i in range(self.shape[0]):
                 name = self.name + f' [{i}]'
                 vals[name] = OutVal(name=name, ncase=self.ncase, val=self.val[i],
-                                        valmap=self.valmap, ismedian=self.ismedian)
+                                    valmap=self.valmap, ismedian=self.ismedian)
         return vals
